@@ -124,6 +124,9 @@ class FakeSessionClient(base_client.SessionClient):
             munged_url = munged_url.replace(' ', '_')
             munged_url = munged_url.replace('!', '_')
             munged_url = munged_url.replace('@', '_')
+            munged_url = munged_url.replace('%20', '_')
+            munged_url = munged_url.replace('%3A', '_')
+            munged_url = munged_url.replace('%', '_')
             callback = "%s_%s" % (method.lower(), munged_url)
 
         if url is None or callback == "get_http:__nova_api:8774":
@@ -174,6 +177,10 @@ class FakeSessionClient(base_client.SessionClient):
             "text": body,
             "headers": headers,
         })
+
+        if status >= 400:
+            raise exceptions.from_response(r, body, url, method)
+
         return r, body
 
     def get_versions(self):
@@ -390,14 +397,32 @@ class FakeSessionClient(base_client.SessionClient):
     #
 
     def get_servers(self, **kw):
-        return (200, {}, {"servers": [
+        servers = {"servers": [
             {'id': '1234', 'name': 'sample-server'},
             {'id': '5678', 'name': 'sample-server2'},
             {'id': '9014', 'name': 'help'}
-        ]})
+        ]}
+        if self.api_version >= api_versions.APIVersion('2.69'):
+            # include "partial results" from non-responsive part of
+            # infrastructure.
+            servers['servers'].append(
+                {'id': '9015',
+                 'status': "UNKNOWN",
+                 "links": [
+                     {
+                         "href": "http://fake/v2.1/",
+                         "rel": "self"
+                     },
+                     {
+                         "href": "http://fake",
+                         "rel": "bookmark"
+                     }
+                 ]}
+            )
+        return (200, {}, servers)
 
     def get_servers_detail(self, **kw):
-        return (200, {}, {"servers": [
+        servers = {"servers": [
             {
                 "id": '1234',
                 "name": "sample-server",
@@ -407,7 +432,7 @@ class FakeSessionClient(base_client.SessionClient):
                 },
                 "flavor": {
                     "id": 1,
-                    "name": "256 MB Server",
+                    "name": "256 MiB Server",
                 },
                 "hostId": "e4d909c290d0fb1ca068ffaddf22cbd0",
                 "status": "BUILD",
@@ -448,7 +473,7 @@ class FakeSessionClient(base_client.SessionClient):
                 },
                 "flavor": {
                     "id": 1,
-                    "name": "256 MB Server",
+                    "name": "256 MiB Server",
                 },
                 "hostId": "9e107d9d372bb6826bd81d3542a419d6",
                 "status": "ACTIVE",
@@ -489,7 +514,7 @@ class FakeSessionClient(base_client.SessionClient):
                 "image": "",
                 "flavor": {
                     "id": 1,
-                    "name": "256 MB Server",
+                    "name": "256 MiB Server",
                 },
                 "hostId": "9e107d9d372bb6826bd81d3542a419d6",
                 "status": "ACTIVE",
@@ -536,7 +561,29 @@ class FakeSessionClient(base_client.SessionClient):
                 "hostId": "9e107d9d372bb6826bd81d3542a419d6",
                 "status": "ACTIVE",
             },
-        ]})
+        ]}
+        if self.api_version >= api_versions.APIVersion('2.69'):
+            # include "partial results" from non-responsive part of
+            # infrastructure.
+            servers['servers'].append(
+                {
+                    "id": "9015",
+                    "status": "UNKNOWN",
+                    "tenant_id": "6f70656e737461636b20342065766572",
+                    "created": "2018-12-03T21:06:18Z",
+                    "links": [
+                        {
+                            "href": "http://fake/v2.1/",
+                            "rel": "self"
+                        },
+                        {
+                            "href": "http://fake",
+                            "rel": "bookmark"
+                        }
+                    ]
+                }
+            )
+        return (200, {}, servers)
 
     def post_servers(self, body, **kw):
         assert set(body.keys()) <= set(['server', 'os:scheduler_hints'])
@@ -571,8 +618,11 @@ class FakeSessionClient(base_client.SessionClient):
         return (202, {}, self.get_servers_9012()[2])
 
     def get_servers_1234(self, **kw):
-        r = {'server': self.get_servers_detail()[2]['servers'][0]}
-        return (200, {}, r)
+        server = self.get_servers_detail()[2]['servers'][0]
+        if self.api_version >= api_versions.APIVersion('2.71'):
+            server.update(
+                {'server_groups': ['a67359fb-d397-4697-88f1-f55e3ee7c499']})
+        return (200, {}, {'server': server})
 
     def get_servers_1235(self, **kw):
         r = {'server': self.get_servers_detail()[2]['servers'][0]}
@@ -597,6 +647,27 @@ class FakeSessionClient(base_client.SessionClient):
         r = {'server': self.get_servers_detail()[2]['servers'][4]}
         return (200, {}, r)
 
+    def get_servers_9015(self, **kw):
+        r = {'server': self.get_servers_detail()[2]['servers'][5]}
+        r['server']["OS-EXT-AZ:availability_zone"] = 'geneva'
+        r['server']["OS-EXT-STS:power_state"] = 0
+        flavor = {
+            "disk": 1,
+            "ephemeral": 0,
+            "original_name": "m1.tiny",
+            "ram": 512,
+            "swap": 0,
+            "vcpus": 1,
+            "extra_specs": {}
+        }
+        image = {
+            "id": "c99d7632-bd66-4be9-aed5-3dd14b223a76",
+        }
+        r['server']['image'] = image
+        r['server']['flavor'] = flavor
+        r['server']['user_id'] = "fake"
+        return (200, {}, r)
+
     def delete_os_server_groups_12345(self, **kw):
         return (202, {}, None)
 
@@ -633,6 +704,12 @@ class FakeSessionClient(base_client.SessionClient):
     def post_servers_uuid4_metadata(self, **kw):
         return (204, {}, {'metadata': {'key1': 'val1'}})
 
+    def post_servers_uuid5_metadata(self, **kw):
+        return (204, {}, {'metadata': {'key1': 'val1'}})
+
+    def post_servers_uuid6_metadata(self, **kw):
+        return (204, {}, {'metadata': {'key1': 'val1'}})
+
     def delete_servers_uuid1_metadata_key1(self, **kw):
         return (200, {}, {'data': 'Fake diagnostics'})
 
@@ -645,6 +722,12 @@ class FakeSessionClient(base_client.SessionClient):
     def delete_servers_uuid4_metadata_key1(self, **kw):
         return (200, {}, {'data': 'Fake diagnostics'})
 
+    def delete_servers_uuid5_metadata_key1(self, **kw):
+        return (200, {}, {'data': 'Fake diagnostics'})
+
+    def delete_servers_uuid6_metadata_key1(self, **kw):
+        return (200, {}, {'data': 'Fake diagnostics'})
+
     def get_servers_1234_os_security_groups(self, **kw):
         return (200, {}, {
             "security_groups": [{
@@ -819,6 +902,8 @@ class FakeSessionClient(base_client.SessionClient):
         included_fields = ['id', 'name']
         if self.api_version >= api_versions.APIVersion('2.55'):
             included_fields.append('description')
+        if self.api_version >= api_versions.APIVersion('2.61'):
+            included_fields.append('extra_specs')
         for flavor in flavors['flavors']:
             for k in list(flavor):
                 if k not in included_fields:
@@ -828,19 +913,19 @@ class FakeSessionClient(base_client.SessionClient):
 
     def get_flavors_detail(self, **kw):
         flavors = {'flavors': [
-            {'id': 1, 'name': '256 MB Server', 'ram': 256, 'disk': 10,
+            {'id': 1, 'name': '256 MiB Server', 'ram': 256, 'disk': 10,
              'OS-FLV-EXT-DATA:ephemeral': 10,
              'os-flavor-access:is_public': True,
              'links': {}},
-            {'id': 2, 'name': '512 MB Server', 'ram': 512, 'disk': 20,
+            {'id': 2, 'name': '512 MiB Server', 'ram': 512, 'disk': 20,
              'OS-FLV-EXT-DATA:ephemeral': 20,
              'os-flavor-access:is_public': False,
              'links': {}},
-            {'id': 4, 'name': '1024 MB Server', 'ram': 1024, 'disk': 10,
+            {'id': 4, 'name': '1024 MiB Server', 'ram': 1024, 'disk': 10,
              'OS-FLV-EXT-DATA:ephemeral': 10,
              'os-flavor-access:is_public': True,
              'links': {}},
-            {'id': 'aa1', 'name': '128 MB Server', 'ram': 128, 'disk': 0,
+            {'id': 'aa1', 'name': '128 MiB Server', 'ram': 128, 'disk': 0,
              'OS-FLV-EXT-DATA:ephemeral': 0,
              'os-flavor-access:is_public': True,
              'links': {}}
@@ -879,6 +964,11 @@ class FakeSessionClient(base_client.SessionClient):
             new_flavor['description'] = 'test description'
             flavors['flavors'].append(new_flavor)
 
+        # Add extra_specs in the response for all flavors.
+        if self.api_version >= api_versions.APIVersion('2.61'):
+            for flavor in flavors['flavors']:
+                flavor['extra_specs'] = {'test': 'value'}
+
         return (200, FAKE_RESPONSE_HEADERS, flavors)
 
     def get_flavors_1(self, **kw):
@@ -904,16 +994,16 @@ class FakeSessionClient(base_client.SessionClient):
             FAKE_RESPONSE_HEADERS,
             {'flavor': {
                 'id': 3,
-                'name': '256 MB Server',
+                'name': '256 MiB Server',
                 'ram': 256,
                 'disk': 10,
             }},
         )
 
-    def get_flavors_512_MB_Server(self, **kw):
+    def get_flavors_512_MiB_Server(self, **kw):
         raise exceptions.NotFound('404')
 
-    def get_flavors_128_MB_Server(self, **kw):
+    def get_flavors_128_MiB_Server(self, **kw):
         raise exceptions.NotFound('404')
 
     def get_flavors_80645cf4_6ad3_410a_bbc8_6f3e1e291f51(self, **kw):
@@ -1623,24 +1713,35 @@ class FakeSessionClient(base_client.SessionClient):
         else:
             service_id_1 = 1
             service_id_2 = 2
-        return (200, FAKE_RESPONSE_HEADERS,
-                {'services': [{'binary': binary,
-                               'host': host,
-                               'zone': 'nova',
-                               'status': 'enabled',
-                               'state': 'up',
-                               'updated_at': datetime.datetime(
-                                   2012, 10, 29, 13, 42, 2),
-                               'id': service_id_1},
-                              {'binary': binary,
-                               'host': host,
-                               'zone': 'nova',
-                               'status': 'disabled',
-                               'state': 'down',
-                               'updated_at': datetime.datetime(
-                                   2012, 9, 18, 8, 3, 38),
-                               'id': service_id_2},
-                              ]})
+        services = {
+            'services': [
+                {'binary': binary,
+                 'host': host,
+                 'zone': 'nova',
+                 'status': 'enabled',
+                 'state': 'up',
+                 'updated_at': datetime.datetime(
+                     2012, 10, 29, 13, 42, 2),
+                 'id': service_id_1},
+                {'binary': binary,
+                 'host': host,
+                 'zone': 'nova',
+                 'status': 'disabled',
+                 'state': 'down',
+                 'updated_at': datetime.datetime(
+                     2012, 9, 18, 8, 3, 38),
+                 'id': service_id_2},
+            ]
+        }
+        if self.api_version >= api_versions.APIVersion('2.69'):
+            services['services'].append(
+                {
+                    "binary": "nova-compute",
+                    "host": "host-down",
+                    "status": "UNKNOWN"
+                }
+            )
+        return (200, FAKE_RESPONSE_HEADERS, services)
 
     def put_os_services_enable(self, body, **kw):
         return (200, FAKE_RESPONSE_HEADERS,
@@ -1773,6 +1874,26 @@ class FakeSessionClient(base_client.SessionClient):
                      {'name': 'inst4', 'uuid': 'uuid4'}]}]
         })
 
+    def get_os_hypervisors_hyper1_servers(self, **kw):
+        return (200, {}, {
+            'hypervisors': [
+                {'id': 1234,
+                 'hypervisor_hostname': 'hyper1',
+                 'servers': [
+                     {'name': 'inst1', 'uuid': 'uuid1'},
+                     {'name': 'inst2', 'uuid': 'uuid2'}]}]
+        })
+
+    def get_os_hypervisors_hyper2_servers(self, **kw):
+        return (200, {}, {
+            'hypervisors': [
+                {'id': 5678,
+                 'hypervisor_hostname': 'hyper2',
+                 'servers': [
+                     {'name': 'inst3', 'uuid': 'uuid3'},
+                     {'name': 'inst4', 'uuid': 'uuid4'}]}]
+        })
+
     def get_os_hypervisors_hyper_no_servers_servers(self, **kw):
         return (200, {}, {'hypervisors':
                           [{'id': 1234, 'hypervisor_hostname': 'hyper1'}]})
@@ -1886,7 +2007,7 @@ class FakeSessionClient(base_client.SessionClient):
                  "hosts": None}]})
 
     def get_servers_1234_os_interface(self, **kw):
-        return (200, {}, {
+        attachments = {
             "interfaceAttachments": [
                 {"port_state": "ACTIVE",
                  "net_id": "net-id-1",
@@ -1899,27 +2020,38 @@ class FakeSessionClient(base_client.SessionClient):
                  "port_id": "port-id-1",
                  "mac_address": "aa:bb:cc:dd:ee:ff",
                  "fixed_ips": [{"ip_address": "1.2.3.4"}],
-                 }]
-        })
+                 }
+            ]
+        }
+        if self.api_version >= api_versions.APIVersion('2.70'):
+            # Include the "tag" field in each attachment.
+            for attachment in attachments['interfaceAttachments']:
+                attachment['tag'] = 'test-tag'
+        return (200, {}, attachments)
 
     def post_servers_1234_os_interface(self, **kw):
-        return (200, {}, {'interfaceAttachment': {}})
+        attachment = {}
+        if self.api_version >= api_versions.APIVersion('2.70'):
+            # Include the "tag" field in the response.
+            attachment['tag'] = 'test-tag'
+        return (200, {}, {'interfaceAttachment': attachment})
 
     def delete_servers_1234_os_interface_port_id(self, **kw):
         return (200, {}, None)
 
     def post_servers_1234_os_volume_attachments(self, **kw):
-        return (200, FAKE_RESPONSE_HEADERS, {
-            "volumeAttachment":
-                {"device": "/dev/vdb",
-                 "volumeId": 2}})
+        attachment = {"device": "/dev/vdb", "volumeId": 2}
+        if self.api_version >= api_versions.APIVersion('2.70'):
+            # Include the "tag" field in the response.
+            attachment['tag'] = 'test-tag'
+        return (200, FAKE_RESPONSE_HEADERS, {"volumeAttachment": attachment})
 
     def put_servers_1234_os_volume_attachments_Work(self, **kw):
         return (200, FAKE_RESPONSE_HEADERS,
                 {"volumeAttachment": {"volumeId": 2}})
 
     def get_servers_1234_os_volume_attachments(self, **kw):
-        return (200, FAKE_RESPONSE_HEADERS, {
+        attachments = {
             "volumeAttachments": [
                 {"display_name": "Work",
                  "display_description": "volume for work",
@@ -1929,7 +2061,14 @@ class FakeSessionClient(base_client.SessionClient):
                  "attached": "2011-11-11T00:00:00Z",
                  "size": 1024,
                  "attachments": [{"id": "3333", "links": ''}],
-                 "metadata": {}}]})
+                 "metadata": {}}
+            ]
+        }
+        if self.api_version >= api_versions.APIVersion('2.70'):
+            # Include the "tag" field in each attachment.
+            for attachment in attachments['volumeAttachments']:
+                attachment['tag'] = 'test-tag'
+        return (200, FAKE_RESPONSE_HEADERS, attachments)
 
     def get_servers_1234_os_volume_attachments_Work(self, **kw):
         return (200, FAKE_RESPONSE_HEADERS, {
@@ -1974,6 +2113,89 @@ class FakeSessionClient(base_client.SessionClient):
         return (200, FAKE_RESPONSE_HEADERS, {
             "instanceAction": action})
 
+    def get_os_instance_usage_audit_log(self, **kw):
+        return (200, FAKE_RESPONSE_HEADERS, {
+            "instance_usage_audit_logs": {
+                "hosts_not_run": ["samplehost3"],
+                "log": {
+                    "samplehost0": {
+                        "errors": 1,
+                        "instances": 1,
+                        "message": ("Instance usage audit ran for host "
+                                    "samplehost0, 1 instances in 0.01 "
+                                    "seconds."),
+                        "state": "DONE"
+                    },
+                    "samplehost1": {
+                        "errors": 1,
+                        "instances": 2,
+                        "message": ("Instance usage audit ran for host "
+                                    "samplehost1, 2 instances in 0.01 "
+                                    "seconds."),
+                        "state": "DONE"
+                    },
+                    "samplehost2": {
+                        "errors": 1,
+                        "instances": 3,
+                        "message": ("Instance usage audit ran for host "
+                                    "samplehost2, 3 instances in 0.01 "
+                                    "seconds."),
+                        "state": "DONE"
+                    },
+                },
+                "num_hosts": 4,
+                "num_hosts_done": 3,
+                "num_hosts_not_run": 1,
+                "num_hosts_running": 0,
+                "overall_status": "3 of 4 hosts done. 3 errors.",
+                "period_beginning": "2012-06-01 00:00:00",
+                "period_ending": "2012-07-01 00:00:00",
+                "total_errors": 3,
+                "total_instances": 6}})
+
+    def get_os_instance_usage_audit_log_2016_12_10_13_59_59_999999(self, **kw):
+        return (200, FAKE_RESPONSE_HEADERS, {
+            "instance_usage_audit_log": {
+                "hosts_not_run": ["samplehost3"],
+                "log": {
+                    "samplehost0": {
+                        "errors": 1,
+                        "instances": 1,
+                        "message": ("Instance usage audit ran for host "
+                                    "samplehost0, 1 instances in 0.01 "
+                                    "seconds."),
+                        "state": "DONE"
+                    },
+                    "samplehost1": {
+                        "errors": 1,
+                        "instances": 2,
+                        "message": ("Instance usage audit ran for host "
+                                    "samplehost1, 2 instances in 0.01 "
+                                    "seconds."),
+                        "state": "DONE"
+                    },
+                    "samplehost2": {
+                        "errors": 1,
+                        "instances": 3,
+                        "message": ("Instance usage audit ran for host "
+                                    "samplehost2, 3 instances in 0.01 "
+                                    "seconds."),
+                        "state": "DONE"
+                    },
+                },
+                "num_hosts": 4,
+                "num_hosts_done": 3,
+                "num_hosts_not_run": 1,
+                "num_hosts_running": 0,
+                "overall_status": "3 of 4 hosts done. 3 errors.",
+                "period_beginning": "2012-06-01 00:00:00",
+                "period_ending": "2012-07-01 00:00:00",
+                "total_errors": 3,
+                "total_instances": 6}})
+
+    def get_os_instance_usage_audit_log__5Cu5de5_5Cu4f5c(self, **kw):
+        return (400, {}, {"Invalid timestamp for date \u6f22\u5b57"})
+
     def post_servers_uuid1_action(self, **kw):
         return 202, {}, {}
 
@@ -1986,6 +2208,12 @@ class FakeSessionClient(base_client.SessionClient):
     def post_servers_uuid4_action(self, **kw):
         return 202, {}, {}
 
+    def post_servers_uuid5_action(self, **kw):
+        return 202, {}, {}
+
+    def post_servers_uuid6_action(self, **kw):
+        return 202, {}, {}
+
     def get_os_cells_child_cell(self, **kw):
         cell = {'cell': {
             'username': 'cell1_user',
@@ -2109,8 +2337,13 @@ class FakeSessionClient(base_client.SessionClient):
         return (200, {}, {"server_groups": server_groups})
 
     def _return_server_group(self):
-        r = {'server_group':
-             self.get_os_server_groups()[2]['server_groups'][0]}
+        if self.api_version < api_versions.APIVersion("2.64"):
+            r = {'server_group':
+                 self.get_os_server_groups()[2]['server_groups'][0]}
+        else:
+            r = {"members": [], "id": "2cbd51f4-fafe-4cdb-801b-cf913a6f288b",
+                 'server_group': {'name': 'ig1', 'policy': 'anti-affinity',
+                                  'rules': {'max_server_per_host': 3}}}
         return (200, {}, r)
 
     def post_os_server_groups(self, body, **kw):
